from rpyc import Connection, connect

from bus_station.passengers.serialization.passenger_serializer import PassengerSerializer
from bus_station.query_terminal.bus.query_bus import QueryBus
from bus_station.query_terminal.handler_not_found_for_query import HandlerNotFoundForQuery
from bus_station.query_terminal.query import Query
from bus_station.query_terminal.query_response import QueryResponse
from bus_station.query_terminal.registry.remote_query_registry import RemoteQueryRegistry
from bus_station.query_terminal.serialization.query_response_deserializer import QueryResponseDeserializer


class RPyCQueryBus(QueryBus):
    def __init__(
        self,
        query_serializer: PassengerSerializer,
        query_response_deserializer: QueryResponseDeserializer,
        query_registry: RemoteQueryRegistry,
    ):
        self.__query_serializer = query_serializer
        self.__query_response_deserializer = query_response_deserializer
        self.__query_registry = query_registry

    def transport(self, passenger: Query) -> QueryResponse:
        query_handler_addr = self.__query_registry.get_query_destination_contact(passenger.__class__)
        if query_handler_addr is None:
            raise HandlerNotFoundForQuery(passenger.passenger_name())

        rpyc_client = self.__get_rpyc_client(query_handler_addr)
        query_response = self.__execute_query(rpyc_client, passenger)
        rpyc_client.close()
        return query_response

    def __get_rpyc_client(self, handler_addr: str) -> Connection:
        host, port = handler_addr.split(":")
        return connect(host, port=port)

    def __execute_query(self, client: Connection, query: Query) -> QueryResponse:
        serialized_query = self.__query_serializer.serialize(query)
        serialized_query_response = getattr(client.root, query.passenger_name())(serialized_query)
        return self.__query_response_deserializer.deserialize(serialized_query_response)



from typing import ClassVar, Type, TypeVar

from kombu import Connection
from kombu.messaging import Exchange, Queue
from kombu.transport.virtual import Channel

from bus_station.command_terminal.command import Command
from bus_station.command_terminal.command_handler import CommandHandler
from bus_station.command_terminal.contact_not_found_for_command import ContactNotFoundForCommand
from bus_station.command_terminal.handler_not_found_for_command import HandlerNotFoundForCommand
from bus_station.command_terminal.registry.remote_command_registry import RemoteCommandRegistry
from bus_station.passengers.passenger_kombu_consumer import PassengerKombuConsumer
from bus_station.passengers.reception.passenger_receiver import PassengerReceiver
from bus_station.passengers.serialization.passenger_deserializer import PassengerDeserializer
from bus_station.shared_terminal.engine.engine import Engine

C = TypeVar("C", bound=Command)


class KombuCommandBusEngine(Engine):
    __DEAD_LETTER_EXCHANGE_NAME: ClassVar = "failed_commands"

    def __init__(
        self,
        broker_connection: Connection,
        command_registry: RemoteCommandRegistry,
        command_receiver: PassengerReceiver[Command, CommandHandler],
        command_deserializer: PassengerDeserializer,
        command_type: Type[C],
    ):
        super().__init__()
        command_handler = command_registry.get_command_destination(command_type)
        if command_handler is None:
            raise HandlerNotFoundForCommand(command_type.passenger_name())

        broker_connection = broker_connection
        channel = broker_connection.channel()
        self.__create_dead_letter_exchange(channel)

        command_queue_name = command_registry.get_command_destination_contact(command_type)
        if command_queue_name is None:
            raise ContactNotFoundForCommand(command_type.passenger_name())

        command_queue = self.__create_command_handler_queue(command_queue_name, channel)

        self.__command_consumer = PassengerKombuConsumer(
            broker_connection,
            command_queue,
            command_handler,
            command_type,
            command_receiver,
            command_deserializer,
        )

    def start(self) -> None:
        try:
            self.__command_consumer.run()
        except KeyboardInterrupt:
            self.stop()

    def stop(self) -> None:
        self.__command_consumer.stop()

    def __create_dead_letter_exchange(self, channel: Channel) -> None:
        command_failure_exchange = Exchange(self.__DEAD_LETTER_EXCHANGE_NAME, type="fanout", channel=channel)
        command_failure_exchange.declare()

    def __create_command_handler_queue(self, command_queue_name: str, channel: Channel) -> Queue:
        command_queue = Queue(
            name=command_queue_name, queue_arguments={"x-dead-letter-exchange": self.__DEAD_LETTER_EXCHANGE_NAME}
        )
        command_queue.declare(channel=channel)
        return command_queue



from abc import abstractmethod
from typing import Optional

from bus_station.command_terminal.command import Command
from bus_station.command_terminal.command_handler import CommandHandler
from bus_station.passengers.passenger_middleware import PassengerMiddleware


class CommandMiddleware(PassengerMiddleware):
    @abstractmethod
    def before_handle(self, passenger: Command, bus_stop: CommandHandler) -> None:
        pass

    @abstractmethod
    def after_handle(
        self, passenger: Command, bus_stop: CommandHandler, handling_exception: Optional[Exception] = None
    ) -> None:
        pass



from bus_station.shared_terminal.engine.runner.engine_runner import EngineRunner


class SelfProcessEngineRunner(EngineRunner):
    def run(self) -> None:
        self._engine.start()

    def stop(self) -> None:
        self._engine.stop()



from typing import Type

from bus_station.tracking_terminal.models.query_tracking import QueryTracking
from bus_station.tracking_terminal.repositories.implementations.pymongo.pymongo_passenger_tracking_repository import (
    PyMongoPassengerTrackingRepository,
)
from bus_station.tracking_terminal.repositories.query_tracking_repository import QueryTrackingRepository


class PyMongoQueryTrackingRepository(PyMongoPassengerTrackingRepository, QueryTrackingRepository):
    @property
    def model(self) -> Type[QueryTracking]:
        return QueryTracking

    @property
    def _collection(self) -> str:
        return "query_tracking"



from bus_station.passengers.reception.passenger_middleware_receiver import PassengerMiddlewareReceiver
from bus_station.query_terminal.middleware.query_middleware import QueryMiddleware
from bus_station.query_terminal.query import Query
from bus_station.query_terminal.query_handler import QueryHandler
from bus_station.query_terminal.query_response import QueryResponse


class QueryMiddlewareReceiver(PassengerMiddlewareReceiver[Query, QueryHandler, QueryMiddleware]):
    def receive(self, passenger: Query, passenger_bus_stop: QueryHandler) -> QueryResponse:
        middlewares = list(self._get_middlewares())
        for middleware in middlewares:
            middleware.before_handle(passenger, passenger_bus_stop)

        query_response = QueryResponse(data=None)
        handling_exception = None
        try:
            query_response = passenger_bus_stop.handle(passenger)
        except Exception as ex:
            handling_exception = ex

        for middleware in reversed(middlewares):
            query_response = middleware.after_handle(
                passenger, passenger_bus_stop, query_response, handling_exception=handling_exception
            )

        if handling_exception is not None:
            raise handling_exception

        return query_response



from dataclasses import asdict
from json import dumps

from bus_station.passengers.passenger import Passenger
from bus_station.passengers.serialization.passenger_serializer import PassengerSerializer


class PassengerJSONSerializer(PassengerSerializer):
    def serialize(self, passenger: Passenger) -> str:
        serialized_data = {
            "passenger_data": self.__get_passenger_data(passenger),
            "passenger_type": self.__get_passenger_type(passenger),
            "passenger_name": passenger.passenger_name(),
        }
        return dumps(serialized_data)

    def __get_passenger_data(self, passenger: Passenger) -> dict:
        return asdict(passenger)

    def __get_passenger_type(self, passenger: Passenger) -> str:
        return ".".join([passenger.__module__, passenger.__class__.__name__])



from abc import abstractmethod

from bus_station.command_terminal.command import Command
from bus_station.shared_terminal.bus_stop import BusStop


class CommandHandler(BusStop):
    @abstractmethod
    def handle(self, command: Command) -> None:
        pass

    @classmethod
    def bus_stop_name(cls) -> str:
        return f"command_handler.{cls.__module__}.{cls.__name__}"



from abc import abstractmethod

from bus_station.query_terminal.query import Query
from bus_station.query_terminal.query_response import QueryResponse
from bus_station.shared_terminal.bus_stop import BusStop


class QueryHandler(BusStop):
    @abstractmethod
    def handle(self, query: Query) -> QueryResponse:
        pass

    @classmethod
    def bus_stop_name(cls) -> str:
        return f"query_handler.{cls.__module__}.{cls.__name__}"



from typing import Optional, Type, TypeVar

from bus_station.command_terminal.command import Command
from bus_station.command_terminal.handler_not_found_for_command import HandlerNotFoundForCommand
from bus_station.command_terminal.json_rpc_command_server import JsonRPCCommandServer
from bus_station.command_terminal.registry.remote_command_registry import RemoteCommandRegistry
from bus_station.shared_terminal.engine.engine import Engine

C = TypeVar("C", bound=Command)


class JsonRPCCommandBusEngine(Engine):
    def __init__(
        self,
        server: JsonRPCCommandServer,
        command_registry: RemoteCommandRegistry,
        command_type: Optional[Type[C]] = None,
    ):
        super().__init__()
        self.__server = server
        if command_type is not None:
            self.__register_command_in_server(command_registry, command_type)
            return

        for command_type in command_registry.get_commands_registered():
            self.__register_command_in_server(command_registry, command_type)

    def __register_command_in_server(self, command_registry: RemoteCommandRegistry, command_type: Type[C]) -> None:
        handler = command_registry.get_command_destination(command_type)
        if handler is None:
            raise HandlerNotFoundForCommand(command_type.passenger_name())
        self.__server.register(command_type, handler)

    def start(self) -> None:
        self.__server.run()

    def stop(self):
        pass



from logging import Logger
from typing import Optional

from bus_station.query_terminal.middleware.query_middleware import QueryMiddleware
from bus_station.query_terminal.query import Query
from bus_station.query_terminal.query_handler import QueryHandler
from bus_station.query_terminal.query_response import QueryResponse


class LoggingQueryMiddleware(QueryMiddleware):
    def __init__(self, logger: Logger):
        self.__logger = logger

    def before_handle(self, passenger: Query, bus_stop: QueryHandler) -> None:
        self.__logger.info(f"Starting handling query {passenger} with {bus_stop.bus_stop_name()}")

    def after_handle(
        self,
        passenger: Query,
        bus_stop: QueryHandler,
        query_response: QueryResponse,
        handling_exception: Optional[Exception] = None,
    ) -> QueryResponse:
        if handling_exception is not None:
            self.__logger.exception(
                handling_exception,
                exc_info=(handling_exception.__class__, handling_exception, handling_exception.__traceback__),
            )
        self.__logger.info(
            f"Finished handling query {passenger} with {bus_stop.bus_stop_name()} with response: {query_response}"
        )
        return query_response



from typing import Type

from bus_station.tracking_terminal.models.event_tracking import EventTracking
from bus_station.tracking_terminal.repositories.event_tracking_repository import EventTrackingRepository
from bus_station.tracking_terminal.repositories.implementations.pymongo.pymongo_passenger_tracking_repository import (
    PyMongoPassengerTrackingRepository,
)


class PyMongoEventTrackingRepository(PyMongoPassengerTrackingRepository, EventTrackingRepository):
    @property
    def model(self) -> Type[EventTracking]:
        return EventTracking

    @property
    def _collection(self) -> str:
        return "event_tracking"



from abc import ABCMeta, abstractmethod
from typing import Any, Iterable, Optional, Type, get_type_hints

from bus_station.query_terminal.handler_for_query_already_registered import HandlerForQueryAlreadyRegistered
from bus_station.query_terminal.query import Query
from bus_station.query_terminal.query_handler import QueryHandler


class QueryRegistry(metaclass=ABCMeta):
    def register(self, handler: QueryHandler, handler_contact: Any) -> None:
        handler_query = self.__get_handler_query(handler)
        existing_handler_contact = self.get_query_destination_contact(handler_query)
        self.__check_query_already_registered(handler_query, handler_contact, existing_handler_contact)
        if existing_handler_contact is None:
            self._register(handler_query, handler, handler_contact)

    def __check_query_already_registered(
        self, query: Type[Query], handler_contact: Any, existing_handler_contact: Optional[Any]
    ) -> None:
        if existing_handler_contact is not None and handler_contact != existing_handler_contact:
            raise HandlerForQueryAlreadyRegistered(query.passenger_name())

    def __get_handler_query(self, handler: QueryHandler) -> Type[Query]:
        handle_typing = get_type_hints(handler.handle)

        if "query" not in handle_typing:
            raise TypeError(f"Handle query not found for {handler.bus_stop_name()}")

        if not issubclass(handle_typing["query"], Query):
            raise TypeError(f"Wrong type for handle query of {handler.bus_stop_name()}")

        return handle_typing["query"]

    @abstractmethod
    def _register(self, query: Type[Query], handler: QueryHandler, handler_contact: Any) -> None:
        pass

    @abstractmethod
    def get_query_destination(self, query: Type[Query]) -> Optional[QueryHandler]:
        pass

    @abstractmethod
    def get_query_destination_contact(self, query: Type[Query]) -> Optional[Any]:
        pass

    @abstractmethod
    def get_queries_registered(self) -> Iterable[Type[Query]]:
        pass

    @abstractmethod
    def unregister(self, query: Type[Query]) -> None:
        pass



class HandlerNotFoundForCommand(Exception):
    def __init__(self, command_name: str):
        self.command_name = command_name
        super(HandlerNotFoundForCommand, self).__init__(f"Handler for command {command_name} not found")



from dataclasses import dataclass, field
from typing import Optional

from bus_station.tracking_terminal.models.passenger_tracking import PassengerTracking


@dataclass
class QueryTracking(PassengerTracking):
    response_data: Optional[dict] = field(default_factory=lambda: None)



import requests
from jsonrpcclient import request
from jsonrpcclient.responses import Error, to_result

from bus_station.passengers.serialization.passenger_serializer import PassengerSerializer
from bus_station.query_terminal.bus.query_bus import QueryBus
from bus_station.query_terminal.handler_not_found_for_query import HandlerNotFoundForQuery
from bus_station.query_terminal.query import Query
from bus_station.query_terminal.query_execution_failed import QueryExecutionFailed
from bus_station.query_terminal.query_response import QueryResponse
from bus_station.query_terminal.registry.remote_query_registry import RemoteQueryRegistry
from bus_station.query_terminal.serialization.query_response_deserializer import QueryResponseDeserializer


class JsonRPCQueryBus(QueryBus):
    def __init__(
        self,
        query_serializer: PassengerSerializer,
        query_response_deserializer: QueryResponseDeserializer,
        query_registry: RemoteQueryRegistry,
    ):
        self.__query_serializer = query_serializer
        self.__query_registry = query_registry
        self.__query_response_deserializer = query_response_deserializer

    def transport(self, passenger: Query) -> QueryResponse:
        handler_address = self.__query_registry.get_query_destination_contact(passenger.__class__)
        if handler_address is None:
            raise HandlerNotFoundForQuery(passenger.passenger_name())

        return self.__execute_query(passenger, handler_address)

    def __execute_query(self, query: Query, query_handler_addr: str) -> QueryResponse:
        serialized_query = self.__query_serializer.serialize(query)

        request_response = requests.post(
            query_handler_addr, json=request(query.passenger_name(), params=(serialized_query,))
        )
        json_rpc_response = to_result(request_response.json())

        if isinstance(json_rpc_response, Error):
            raise QueryExecutionFailed(query, json_rpc_response.message)
        return self.__query_response_deserializer.deserialize(json_rpc_response.result)



from abc import abstractmethod
from typing import Optional

from bus_station.passengers.passenger_middleware import PassengerMiddleware
from bus_station.query_terminal.query import Query
from bus_station.query_terminal.query_handler import QueryHandler
from bus_station.query_terminal.query_response import QueryResponse


class QueryMiddleware(PassengerMiddleware):
    @abstractmethod
    def before_handle(self, passenger: Query, bus_stop: QueryHandler) -> None:
        pass

    @abstractmethod
    def after_handle(
        self,
        passenger: Query,
        bus_stop: QueryHandler,
        query_response: QueryResponse,
        handling_exception: Optional[Exception] = None,
    ) -> QueryResponse:
        pass



from typing import Type

from bus_station.passengers.passenger import Passenger


class PassengerDeserializationError(Exception):
    def __init__(self, passenger: Type[Passenger], reason: str):
        self.reason = reason
        self.passenger = passenger
        super().__init__(f"Error deserializing {passenger.__name__}. Reason: {reason}")



import os
import signal
from multiprocessing import Process

from bus_station.shared_terminal.engine.engine import Engine
from bus_station.shared_terminal.engine.runner.engine_runner import EngineRunner


class ProcessEngineRunner(EngineRunner):
    def __init__(self, engine: Engine, should_interrupt: bool):
        super().__init__(engine)
        self.__process = Process(target=engine.start)
        self.__should_interrupt = should_interrupt

    def run(self) -> None:
        self.__process.start()

    def stop(self) -> None:
        self._engine.stop()
        if self.__process.is_alive() is False:
            return

        if self.__should_interrupt is True and (process_pid := self.__process.pid) is not None:
            os.kill(process_pid, signal.SIGINT)
        self.__process.join()



from dataclasses import asdict
from typing import Any, Type, TypeVar

P = TypeVar("P")


class PassengerTrackingToProxyTransformer:
    def transform(self, tracking_instance: Any, target_proxy_class: Type[P]) -> P:
        target_proxy_data = asdict(tracking_instance)
        return target_proxy_class(**target_proxy_data)



import time
from logging import Logger
from typing import Optional

from bus_station.event_terminal.event import Event
from bus_station.event_terminal.event_consumer import EventConsumer
from bus_station.event_terminal.middleware.event_middleware import EventMiddleware


class TimingEventMiddleware(EventMiddleware):
    def __init__(self, logger: Logger):
        self.__logger = logger
        self.__start_time = None

    def before_consume(self, passenger: Event, bus_stop: EventConsumer) -> None:
        self.__start_time = time.time()

    def after_consume(
        self, passenger: Event, bus_stop: EventConsumer, consume_exception: Optional[Exception] = None
    ) -> None:
        execution_time = time.time() - self.__start_time
        execution_result_hint = "successfully" if consume_exception is None else "wrongly"
        self.__logger.info(
            f"Event {passenger} consumed {execution_result_hint} "
            f"by {bus_stop.bus_stop_name()} in {execution_time} seconds"
        )



from abc import ABC

from bus_station.tracking_terminal.models.command_tracking import CommandTracking
from bus_station.tracking_terminal.repositories.passenger_tracking_repository import PassengerTrackingRepository


class CommandTrackingRepository(PassengerTrackingRepository[CommandTracking], ABC):
    pass



from abc import abstractmethod
from typing import Generic, Optional, TypeVar

from bus_station.tracking_terminal.models.passenger_tracking import PassengerTracking

T = TypeVar("T", bound=PassengerTracking)


class PassengerTrackingRepository(Generic[T]):
    @abstractmethod
    def save(self, passenger_tracking: T) -> None:
        pass

    @abstractmethod
    def find_by_id(self, passenger_tracking_id: str) -> Optional[T]:
        pass



from abc import abstractmethod
from functools import partial
from http.server import ThreadingHTTPServer
from typing import Callable, Dict, Generic, Optional, Type, TypeVar

from jsonrpcserver import Error, Success, method
from jsonrpcserver.codes import ERROR_INTERNAL_ERROR
from jsonrpcserver.result import Result
from jsonrpcserver.server import RequestHandler

from bus_station.passengers.passenger import Passenger
from bus_station.passengers.reception.passenger_receiver import PassengerReceiver
from bus_station.passengers.serialization.passenger_deserializer import PassengerDeserializer
from bus_station.shared_terminal.bus_stop import BusStop

P = TypeVar("P", bound=Passenger)
S = TypeVar("S", bound=BusStop)


class JsonRPCServer(Generic[P, S]):
    def __init__(
        self,
        host: str,
        port: int,
        passenger_deserializer: PassengerDeserializer,
        passenger_receiver: PassengerReceiver[P, S],
    ):
        self.__host = host
        self.__port = port
        self._passenger_deserializer = passenger_deserializer
        self._passenger_receiver = passenger_receiver
        self.__internal_registry: Dict[str, Callable] = {}

    def register(self, passenger_class: Type[P], bus_stop: S) -> None:
        exposed_callable = partial(self.__passenger_handler, bus_stop, passenger_class)
        exposed_callable_name = passenger_class.passenger_name()
        self.__internal_registry[exposed_callable_name] = exposed_callable

    def run(self) -> None:
        self.__register_json_rpc_handlers()

        http_server = ThreadingHTTPServer((self.__host, self.__port), RequestHandler)
        try:
            http_server.serve_forever()
        except KeyboardInterrupt:
            http_server.server_close()

    def __register_json_rpc_handlers(self) -> None:
        for exposed_callable_name, exposed_callable in self.__internal_registry.items():
            method(exposed_callable, name=exposed_callable_name)

    def __passenger_handler(self, bus_stop: S, passenger_class: Type[P], serialized_passenger: str) -> Result:
        try:
            serialized_response = self._passenger_handler(bus_stop, passenger_class, serialized_passenger)
        except Exception as ex:
            return Error(code=ERROR_INTERNAL_ERROR, message=str(ex))
        return Success(result=serialized_response)

    @abstractmethod
    def _passenger_handler(self, bus_stop: S, passenger_class: Type[P], serialized_passenger: str) -> Optional[str]:
        pass



from typing import Type, TypeVar

from bus_station.command_terminal.command import Command
from bus_station.command_terminal.command_handler import CommandHandler
from bus_station.command_terminal.handler_not_found_for_command import HandlerNotFoundForCommand
from bus_station.command_terminal.registry.command_registry import CommandRegistry
from bus_station.passengers.memory_queue_passenger_worker import MemoryQueuePassengerWorker
from bus_station.passengers.reception.passenger_receiver import PassengerReceiver
from bus_station.passengers.serialization.passenger_deserializer import PassengerDeserializer
from bus_station.shared_terminal.engine.engine import Engine

C = TypeVar("C", bound=Command)


class MemoryQueueCommandBusEngine(Engine):
    def __init__(
        self,
        command_registry: CommandRegistry,
        command_receiver: PassengerReceiver[Command, CommandHandler],
        command_deserializer: PassengerDeserializer,
        command_type: Type[C],
    ):
        super().__init__()
        command_handler = command_registry.get_command_destination(command_type)
        if command_handler is None:
            raise HandlerNotFoundForCommand(command_type.passenger_name())

        command_queue = command_registry.get_command_destination_contact(command_type)
        self.__command_worker = MemoryQueuePassengerWorker(
            command_queue, command_handler, command_receiver, command_deserializer
        )

    def start(self) -> None:
        try:
            self.__command_worker.work()
        except KeyboardInterrupt:
            self.stop()

    def stop(self) -> None:
        self.__command_worker.stop()



from logging import Logger
from typing import Optional

from bus_station.event_terminal.event import Event
from bus_station.event_terminal.event_consumer import EventConsumer
from bus_station.event_terminal.middleware.event_middleware import EventMiddleware


class LoggingEventMiddleware(EventMiddleware):
    def __init__(self, logger: Logger):
        self.__logger = logger

    def before_consume(self, passenger: Event, bus_stop: EventConsumer) -> None:
        self.__logger.info(f"Starting consuming event {passenger} with {bus_stop.bus_stop_name()}")

    def after_consume(
        self, passenger: Event, bus_stop: EventConsumer, consume_exception: Optional[Exception] = None
    ) -> None:
        if consume_exception is not None:
            self.__logger.exception(
                consume_exception,
                exc_info=(consume_exception.__class__, consume_exception, consume_exception.__traceback__),
            )
        self.__logger.info(f"Finished consuming event {passenger} with {bus_stop.bus_stop_name()}")



from abc import abstractmethod

from bus_station.event_terminal.event import Event
from bus_station.shared_terminal.bus_stop import BusStop


class EventConsumer(BusStop):
    @abstractmethod
    def consume(self, event: Event) -> None:
        pass

    @classmethod
    def bus_stop_name(cls) -> str:
        return f"event_consumer.{cls.__module__}.{cls.__name__}"



from abc import ABC, abstractmethod
from typing import Optional, Type

from bus_station.command_terminal.command import Command
from bus_station.command_terminal.command_handler import CommandHandler
from bus_station.command_terminal.registry.command_registry import CommandRegistry


class RemoteCommandRegistry(CommandRegistry, ABC):
    @abstractmethod
    def _register(self, command: Type[Command], handler: CommandHandler, handler_contact: str) -> None:
        pass

    @abstractmethod
    def get_command_destination_contact(self, command: Type[Command]) -> Optional[str]:
        pass



from typing import ClassVar

from kombu import Connection
from kombu.messaging import Exchange, Queue
from kombu.transport.virtual import Channel

from bus_station.event_terminal.contact_not_found_for_consumer import ContactNotFoundForConsumer
from bus_station.event_terminal.event import Event
from bus_station.event_terminal.event_consumer import EventConsumer
from bus_station.event_terminal.registry.remote_event_registry import RemoteEventRegistry
from bus_station.passengers.passenger_kombu_consumer import PassengerKombuConsumer
from bus_station.passengers.reception.passenger_receiver import PassengerReceiver
from bus_station.passengers.serialization.passenger_deserializer import PassengerDeserializer
from bus_station.shared_terminal.engine.engine import Engine


class KombuEventBusEngine(Engine):
    __DEAD_LETTER_EXCHANGE_NAME: ClassVar = "failed_events"

    def __init__(
        self,
        broker_connection: Connection,
        event_registry: RemoteEventRegistry,
        event_receiver: PassengerReceiver[Event, EventConsumer],
        event_deserializer: PassengerDeserializer,
        event_consumer: EventConsumer,
    ):
        super().__init__()
        event_exchange_name = event_registry.get_event_destination_contact(event_consumer)
        if event_exchange_name is None:
            raise ContactNotFoundForConsumer(event_consumer.bus_stop_name())

        broker_connection = broker_connection
        channel = broker_connection.channel()
        self.__create_dead_letter_exchange(channel)

        event_exchange = self.__create_event_consumer_exchange(event_exchange_name, channel)
        event_consumer_queue = self.__create_event_consumer_queue(event_consumer, event_exchange, channel)

        self.__event_consumer_consumer = PassengerKombuConsumer(
            broker_connection,
            event_consumer_queue,
            event_consumer,
            event_registry.get_consumer_event(event_consumer),
            event_receiver,
            event_deserializer,
        )

    def start(self) -> None:
        try:
            self.__event_consumer_consumer.run()
        except KeyboardInterrupt:
            self.stop()

    def stop(self) -> None:
        self.__event_consumer_consumer.stop()

    def __create_dead_letter_exchange(self, channel: Channel) -> None:
        command_failure_exchange = Exchange(self.__DEAD_LETTER_EXCHANGE_NAME, type="fanout", channel=channel)
        command_failure_exchange.declare()

    def __create_event_consumer_exchange(self, event_name: str, broker_channel: Channel) -> Exchange:
        event_exchange = Exchange(event_name, type="fanout")
        event_exchange.declare(channel=broker_channel)
        return event_exchange

    def __create_event_consumer_queue(
        self, event_consumer: EventConsumer, event_exchange: Exchange, broker_channel: Channel
    ) -> Queue:
        consumer_queue = Queue(
            name=event_consumer.bus_stop_name(),
            exchange=event_exchange,
            queue_arguments={"x-dead-letter-exchange": self.__DEAD_LETTER_EXCHANGE_NAME},
        )
        consumer_queue.declare(channel=broker_channel)
        return consumer_queue



from abc import abstractmethod

from bus_station.command_terminal.command import Command
from bus_station.shared_terminal.bus import Bus


class CommandBus(Bus[Command]):
    @abstractmethod
    def transport(self, passenger: Command) -> None:
        pass



from dataclasses import asdict
from typing import Optional

from bus_station.query_terminal.middleware.query_middleware import QueryMiddleware
from bus_station.query_terminal.query import Query
from bus_station.query_terminal.query_handler import QueryHandler
from bus_station.query_terminal.query_response import QueryResponse
from bus_station.tracking_terminal.trackers.passenger_tracker import PassengerTracker


class TrackingQueryMiddleware(QueryMiddleware):
    def __init__(self, passenger_tracker: PassengerTracker):
        self.__tracker = passenger_tracker

    def before_handle(self, passenger: Query, bus_stop: QueryHandler) -> None:
        self.__tracker.start_tracking(passenger, bus_stop)

    def after_handle(
        self,
        passenger: Query,
        bus_stop: QueryHandler,
        query_response: QueryResponse,
        handling_exception: Optional[Exception] = None,
    ) -> QueryResponse:
        success = handling_exception is None
        response_data = asdict(query_response) if query_response is not None else None
        self.__tracker.end_tracking(passenger, response_data=response_data, success=success)
        return query_response



from typing import Iterable, List, Optional

from redis import Redis

from bus_station.passengers.passenger_record.passenger_record import PassengerRecord
from bus_station.passengers.passenger_record.passenger_record_repository import PassengerRecordRepository


class RedisPassengerRecordRepository(PassengerRecordRepository):
    def __init__(self, client: Redis):
        self.__client = client

    def save(self, passenger_record: PassengerRecord[str]) -> None:
        str_passenger_record = str(passenger_record)
        self.__client.lpush(passenger_record.passenger_name, str_passenger_record.encode("UTF-8"))

    def find_by_passenger_name(self, passenger_name: str) -> Optional[List[PassengerRecord]]:
        redis_records = []
        for redis_value in self.__client.lrange(passenger_name, 0, -1):
            passenger_record = PassengerRecord.from_str(redis_value.decode("UTF-8"))
            redis_records.append(passenger_record)

        if len(redis_records) == 0:
            return None
        return redis_records

    def find_by_passenger_name_and_destination(
        self, passenger_name: str, passenger_destination_fqn: str
    ) -> Optional[PassengerRecord]:
        for redis_value in self.__client.lrange(passenger_name, 0, -1):
            passenger_record = PassengerRecord.from_str(redis_value.decode("UTF-8"))
            if passenger_record.destination_fqn == passenger_destination_fqn:
                return passenger_record
        return None

    def all(self) -> Iterable[PassengerRecord]:
        for passenger_name in self.__client.scan_iter("*"):
            passenger_records = self.find_by_passenger_name(passenger_name)
            if passenger_records is None:
                continue
            for passenger_record in passenger_records:
                yield passenger_record

    def delete_by_passenger_name(self, passenger_name: str) -> None:
        self.__client.delete(passenger_name)



class PassengerTrackingNotFound(Exception):
    def __init__(self, passenger_name: str, passenger_tracking_id: str):
        self.passenger_name = passenger_name
        self.passenger_tracking_id = passenger_tracking_id
        super().__init__(f"Passenger {passenger_name} with tracking id {passenger_tracking_id} not found")



from logging import Logger
from typing import Optional

from bus_station.command_terminal.command import Command
from bus_station.command_terminal.command_handler import CommandHandler
from bus_station.command_terminal.middleware.command_middleware import CommandMiddleware


class LoggingCommandMiddleware(CommandMiddleware):
    def __init__(self, logger: Logger):
        self.__logger = logger

    def before_handle(self, passenger: Command, bus_stop: CommandHandler) -> None:
        self.__logger.info(f"Starting handling command {passenger} with {bus_stop.bus_stop_name()}")

    def after_handle(
        self, passenger: Command, bus_stop: CommandHandler, handling_exception: Optional[Exception] = None
    ) -> None:
        if handling_exception is not None:
            self.__logger.exception(
                handling_exception,
                exc_info=(handling_exception.__class__, handling_exception, handling_exception.__traceback__),
            )
        self.__logger.info(f"Finished handling command {passenger} with {bus_stop.bus_stop_name()}")



from typing import Any, List, Optional, Set, Type

from bus_station.event_terminal.event import Event
from bus_station.event_terminal.event_consumer import EventConsumer
from bus_station.event_terminal.registry.event_registry import EventRegistry
from bus_station.passengers.passenger_class_resolver import PassengerClassResolver
from bus_station.passengers.passenger_record.in_memory_passenger_record_repository import (
    InMemoryPassengerRecordRepository,
)
from bus_station.passengers.passenger_record.passenger_record import PassengerRecord
from bus_station.shared_terminal.bus_stop_resolver.bus_stop_resolver import BusStopResolver
from bus_station.shared_terminal.fqn_getter import FQNGetter


class InMemoryEventRegistry(EventRegistry):
    def __init__(
        self,
        in_memory_repository: InMemoryPassengerRecordRepository,
        fqn_getter: FQNGetter,
        event_consumer_resolver: BusStopResolver[EventConsumer],
        passenger_class_resolver: PassengerClassResolver,
    ):
        self.__in_memory_repository = in_memory_repository
        self.__fqn_getter = fqn_getter
        self.__event_consumer_resolver = event_consumer_resolver
        self.__passenger_class_resolver = passenger_class_resolver

    def _register(self, event: Type[Event], consumer: EventConsumer, consumer_contact: Any) -> None:
        self.__in_memory_repository.save(
            PassengerRecord(
                passenger_name=event.passenger_name(),
                passenger_fqn=self.__fqn_getter.get(event),
                destination_fqn=self.__fqn_getter.get(consumer),
                destination_contact=consumer_contact,
            )
        )

    def get_event_destinations(self, event: Type[Event]) -> Optional[Set[EventConsumer]]:
        event_records: Optional[List[PassengerRecord[Any]]] = self.__in_memory_repository.find_by_passenger_name(
            event.passenger_name()
        )
        if event_records is None:
            return None

        event_destinations = set()
        for event_record in event_records:
            if event_record.destination_fqn is None:
                continue
            event_destinations.add(self.__event_consumer_resolver.resolve_from_fqn(event_record.destination_fqn))
        return event_destinations

    def get_event_destination_contact(self, event_destination: EventConsumer) -> Optional[Any]:
        event = self.get_consumer_event(event_destination)
        event_record = self.__in_memory_repository.find_by_passenger_name_and_destination(
            passenger_name=event.passenger_name(), passenger_destination_fqn=self.__fqn_getter.get(event_destination)
        )
        if event_record is None:
            return None

        return event_record.destination_contact

    def get_event_destination_contacts(self, event: Type[Event]) -> Optional[Set[Any]]:
        event_records: Optional[List[PassengerRecord[Any]]] = self.__in_memory_repository.find_by_passenger_name(
            event.passenger_name()
        )
        if event_records is None:
            return None

        event_destination_contacts = set()
        for event_record in event_records:
            if event_record.destination_contact in event_destination_contacts:
                continue
            event_destination_contacts.add(event_record.destination_contact)
        return event_destination_contacts

    def get_events_registered(self) -> Set[Type[Event]]:
        events_registered = set()
        for event_record in self.__in_memory_repository.all():
            event = self.__passenger_class_resolver.resolve_from_fqn(event_record.passenger_fqn)
            if event is None or not issubclass(event, Event):
                continue
            events_registered.add(event)
        return events_registered

    def unregister(self, event: Type[Event]) -> None:
        self.__in_memory_repository.delete_by_passenger_name(event.passenger_name())



class ContactNotFoundForCommand(Exception):
    def __init__(self, command_name: str):
        self.command_name = command_name
        super().__init__(f"Contact for command {command_name} not found")



from bus_station.query_terminal.query import Query


class QueryExecutionFailed(Exception):
    def __init__(self, query: Query, reason: str):
        self.query = query
        self.reason = reason
        super().__init__(f"{query} execution has failed due to {reason}")



from typing import TypeVar

from bus_station.event_terminal.contact_not_found_for_consumer import ContactNotFoundForConsumer
from bus_station.event_terminal.event import Event
from bus_station.event_terminal.event_consumer import EventConsumer
from bus_station.event_terminal.registry.event_registry import EventRegistry
from bus_station.passengers.memory_queue_passenger_worker import MemoryQueuePassengerWorker
from bus_station.passengers.reception.passenger_receiver import PassengerReceiver
from bus_station.passengers.serialization.passenger_deserializer import PassengerDeserializer
from bus_station.shared_terminal.engine.engine import Engine

E = TypeVar("E", bound=Event)


class MemoryQueueEventBusEngine(Engine):
    def __init__(
        self,
        event_registry: EventRegistry,
        event_receiver: PassengerReceiver[Event, EventConsumer],
        event_deserializer: PassengerDeserializer,
        event_consumer: EventConsumer,
    ):
        super().__init__()
        event_queue = event_registry.get_event_destination_contact(event_consumer)
        if event_queue is None:
            raise ContactNotFoundForConsumer(event_consumer.bus_stop_name())

        self.__event_worker = MemoryQueuePassengerWorker(
            event_queue, event_consumer, event_receiver, event_deserializer
        )

    def start(self) -> None:
        try:
            self.__event_worker.work()
        except KeyboardInterrupt:
            self.stop()

    def stop(self) -> None:
        self.__event_worker.stop()



from typing import Optional

from bus_station.event_terminal.event import Event
from bus_station.event_terminal.event_consumer import EventConsumer
from bus_station.event_terminal.middleware.event_middleware import EventMiddleware
from bus_station.tracking_terminal.trackers.passenger_tracker import PassengerTracker


class TrackingEventMiddleware(EventMiddleware):
    def __init__(self, passenger_tracker: PassengerTracker):
        self.__tracker = passenger_tracker

    def before_consume(self, passenger: Event, bus_stop: EventConsumer) -> None:
        self.__tracker.start_tracking(passenger, bus_stop)

    def after_consume(
        self, passenger: Event, bus_stop: EventConsumer, consume_exception: Optional[Exception] = None
    ) -> None:
        success = consume_exception is None
        self.__tracker.end_tracking(passenger, success)



class HandlerForCommandAlreadyRegistered(Exception):
    def __init__(self, command_name: str):
        self.command_name = command_name
        super(HandlerForCommandAlreadyRegistered, self).__init__(
            f"Handler for command {command_name} already registered"
        )



from typing import Optional

from bus_station.command_terminal.command import Command
from bus_station.command_terminal.command_handler import CommandHandler
from bus_station.command_terminal.middleware.command_middleware import CommandMiddleware
from bus_station.tracking_terminal.trackers.passenger_tracker import PassengerTracker


class TrackingCommandMiddleware(CommandMiddleware):
    def __init__(self, passenger_tracker: PassengerTracker):
        self.__tracker = passenger_tracker

    def before_handle(self, passenger: Command, bus_stop: CommandHandler) -> None:
        self.__tracker.start_tracking(passenger, bus_stop)

    def after_handle(
        self, passenger: Command, bus_stop: CommandHandler, handling_exception: Optional[Exception] = None
    ) -> None:
        success = handling_exception is None
        self.__tracker.end_tracking(passenger, success)



from abc import ABCMeta, abstractmethod
from typing import Any, Iterable, Optional, Type, get_type_hints

from bus_station.command_terminal.command import Command
from bus_station.command_terminal.command_handler import CommandHandler
from bus_station.command_terminal.handler_for_command_already_registered import HandlerForCommandAlreadyRegistered


class CommandRegistry(metaclass=ABCMeta):
    def register(self, handler: CommandHandler, handler_contact: Any) -> None:
        handler_command = self.__get_handler_command(handler)
        existing_handler_contact = self.get_command_destination_contact(handler_command)
        self.__check_command_already_registered(handler_command, handler_contact, existing_handler_contact)
        if existing_handler_contact is None:
            self._register(handler_command, handler, handler_contact)

    def __check_command_already_registered(
        self, command: Type[Command], handler_contact: Any, existing_handler_contact: Optional[Any]
    ) -> None:
        if existing_handler_contact is not None and handler_contact != existing_handler_contact:
            raise HandlerForCommandAlreadyRegistered(command.passenger_name())

    def __get_handler_command(self, handler: CommandHandler) -> Type[Command]:
        handle_typing = get_type_hints(handler.handle)

        if "command" not in handle_typing:
            raise TypeError(f"Handle command not found for {handler.bus_stop_name()}")

        if not issubclass(handle_typing["command"], Command):
            raise TypeError(f"Wrong type for handle command of {handler.bus_stop_name()}")

        return handle_typing["command"]

    @abstractmethod
    def _register(self, command: Type[Command], handler: CommandHandler, handler_contact: Any) -> None:
        pass

    @abstractmethod
    def get_command_destination(self, command: Type[Command]) -> Optional[CommandHandler]:
        pass

    @abstractmethod
    def get_command_destination_contact(self, command: Type[Command]) -> Optional[Any]:
        pass

    @abstractmethod
    def get_commands_registered(self) -> Iterable[Type[Command]]:
        pass

    @abstractmethod
    def unregister(self, command: Type[Command]) -> None:
        pass



from dataclasses import asdict
from typing import Any, Type, TypeVar

T = TypeVar("T")


class ProxyToPassengerTrackingTransformer:
    def transform(self, proxy_instance: Any, target_class: Type[T]) -> T:
        target_instance_data = asdict(proxy_instance)
        return target_class(**target_instance_data)



from abc import abstractmethod
from functools import partial
from typing import Callable, ClassVar, Dict, Generic, Optional, Type, TypeVar

from rpyc import Service, ThreadedServer

from bus_station.passengers.passenger import Passenger
from bus_station.passengers.reception.passenger_receiver import PassengerReceiver
from bus_station.passengers.serialization.passenger_deserializer import PassengerDeserializer
from bus_station.shared_terminal.bus_stop import BusStop

P = TypeVar("P", bound=Passenger)
S = TypeVar("S", bound=BusStop)


class RPyCServer(Generic[P, S]):
    __EXPOSED_CALLABLE_PATTERN: ClassVar[str] = "exposed_{callable_name}"

    def __init__(
        self,
        host: str,
        port: int,
        passenger_deserializer: PassengerDeserializer,
        passenger_receiver: PassengerReceiver,
    ):
        self._passenger_deserializer = passenger_deserializer
        self._passenger_receiver = passenger_receiver
        self.__host = host
        self.__port = port
        self.__callables_to_expose: Dict[str, Callable] = {}

    def register(self, passenger_class: Type[P], bus_stop: S) -> None:
        exposed_callable = partial(self._passenger_handler, bus_stop, passenger_class)
        exposed_callable_name = self.__EXPOSED_CALLABLE_PATTERN.format(callable_name=passenger_class.passenger_name())
        self.__callables_to_expose[exposed_callable_name] = exposed_callable

    def run(self) -> None:
        rpyc_server = ThreadedServer(
            service=self.__build_rpyc_service(),
            hostname=self.__host,
            port=self.__port,
            protocol_config={"allow_public_attrs": True},
        )
        rpyc_server.start()
        rpyc_server.close()

    def __build_rpyc_service(self) -> Service:
        rpyc_service_class = type("RPyCServiceClass", (Service,), self.__callables_to_expose)
        return rpyc_service_class()

    @abstractmethod
    def _passenger_handler(self, bus_stop: S, passenger_class: Type[P], serialized_passenger: str) -> Optional[str]:
        pass



from dataclasses import dataclass

from bus_station.passengers.passenger import Passenger


@dataclass(frozen=True)
class Query(Passenger):
    @classmethod
    def passenger_name(cls) -> str:
        return f"query.{cls.__module__}.{cls.__name__}"



from importlib import import_module
from typing import Optional, Type

from bus_station.passengers.passenger import Passenger


class PassengerClassResolver:
    def resolve_from_fqn(self, passenger_class_fqn: str) -> Optional[Type[Passenger]]:
        class_components = passenger_class_fqn.rsplit(".", 1)
        module_name = class_components[0]
        class_name = class_components[1]

        module = import_module(module_name)
        return getattr(module, class_name)



from typing import Type

from bus_station.passengers.passenger import Passenger
from bus_station.tracking_terminal.repositories.implementations.pymongo.pymongo_passenger_tracking_repository import (
    PyMongoPassengerTrackingRepository,
)
from bus_station.tracking_terminal.trackers.passenger_tracker import PassengerTracker


class PyMongoPassengerTracker(PassengerTracker):
    def __init__(self, passenger_tracking_repository: PyMongoPassengerTrackingRepository):
        super().__init__(passenger_tracking_repository)
        self.tracking_model = passenger_tracking_repository.model

    def _get_tracking_model(self, passenger: Passenger) -> Type:
        return self.tracking_model



from abc import ABC

from bus_station.tracking_terminal.models.query_tracking import QueryTracking
from bus_station.tracking_terminal.repositories.passenger_tracking_repository import PassengerTrackingRepository


class QueryTrackingRepository(PassengerTrackingRepository[QueryTracking], ABC):
    pass



from types import new_class
from typing import ClassVar, Type


class ProxyClassFactory:
    __PROXY_NAME_PATTERN: ClassVar[str] = "Proxy{target_class_name}"

    @classmethod
    def get_proxy(cls, target_class: Type) -> Type:
        proxy_name = cls.__PROXY_NAME_PATTERN.format(target_class_name=target_class.__name__)
        return new_class(proxy_name, bases=(target_class,))



class ContactNotFoundForConsumer(Exception):
    def __init__(self, event_consumer_name: str):
        self.event_consumer_name = event_consumer_name
        super().__init__(f"Contact for event consumer {event_consumer_name} not found")



import json
from dataclasses import asdict

from bus_station.query_terminal.query_response import QueryResponse
from bus_station.query_terminal.serialization.query_response_serializer import QueryResponseSerializer


class QueryResponseJSONSerializer(QueryResponseSerializer):
    def serialize(self, query_response: QueryResponse) -> str:
        return json.dumps(asdict(query_response))



from dataclasses import dataclass
from datetime import datetime
from typing import Optional


@dataclass
class PassengerTracking:
    id: str
    name: str
    executor_name: str
    data: dict
    execution_start: Optional[datetime]
    execution_end: Optional[datetime]
    success: Optional[bool]



from multiprocessing import Queue

from bus_station.event_terminal.bus.event_bus import EventBus
from bus_station.event_terminal.event import Event
from bus_station.event_terminal.registry.in_memory_event_registry import InMemoryEventRegistry
from bus_station.passengers.serialization.passenger_serializer import PassengerSerializer


class MemoryQueueEventBus(EventBus):
    def __init__(
        self,
        event_serializer: PassengerSerializer,
        event_registry: InMemoryEventRegistry,
    ):
        self.__event_serializer = event_serializer
        self.__event_registry = event_registry

    def transport(self, passenger: Event) -> None:
        event_consumer_queues = self.__event_registry.get_event_destination_contacts(passenger.__class__)
        if event_consumer_queues is None:
            return

        for event_queue in event_consumer_queues:
            self.__put_event(event_queue, passenger)

    def __put_event(self, queue: Queue, event: Event) -> None:
        serialized_event = self.__event_serializer.serialize(event)
        queue.put(serialized_event)



from bus_station.command_terminal.command import Command
from bus_station.command_terminal.command_handler import CommandHandler
from bus_station.command_terminal.middleware.command_middleware import CommandMiddleware
from bus_station.passengers.reception.passenger_middleware_receiver import PassengerMiddlewareReceiver


class CommandMiddlewareReceiver(PassengerMiddlewareReceiver[Command, CommandHandler, CommandMiddleware]):
    def receive(self, passenger: Command, passenger_bus_stop: CommandHandler) -> None:
        middlewares = list(self._get_middlewares())
        for middleware in middlewares:
            middleware.before_handle(passenger, passenger_bus_stop)

        handling_exception = None
        try:
            passenger_bus_stop.handle(passenger)
        except Exception as ex:
            handling_exception = ex

        for middleware in reversed(middlewares):
            middleware.after_handle(passenger, passenger_bus_stop, handling_exception=handling_exception)

        if handling_exception is not None:
            raise handling_exception



from abc import ABC

from bus_station.tracking_terminal.models.event_tracking import EventTracking
from bus_station.tracking_terminal.repositories.passenger_tracking_repository import PassengerTrackingRepository


class EventTrackingRepository(PassengerTrackingRepository[EventTracking], ABC):
    pass



from inspect import isclass
from typing import Any, ClassVar


class FQNGetter:

    __FQN_PATTERN: ClassVar[str] = "{module_name}.{class_qualname}"

    def get(self, subject: Any) -> str:
        cls = subject
        if not isclass(cls):
            cls = cls.__class__
        return self.__FQN_PATTERN.format(module_name=cls.__module__, class_qualname=cls.__qualname__)



from typing import Optional, Type, TypeVar

from bus_station.query_terminal.query import Query
from bus_station.query_terminal.registry.query_registry import QueryRegistry
from bus_station.shared_terminal.engine.engine import Engine
from bus_station.shared_terminal.rpyc_server import RPyCServer

Q = TypeVar("Q", bound=Query)


class RPyCQueryBusEngine(Engine):
    def __init__(self, rpyc_server: RPyCServer, query_registry: QueryRegistry, query_type: Optional[Type[Q]] = None):
        super().__init__()
        self.__server = rpyc_server

        if query_type is not None:
            self.__register_query_in_server(query_registry, query_type)
            return

        for command_type in query_registry.get_queries_registered():
            self.__register_query_in_server(query_registry, command_type)

    def __register_query_in_server(self, query_registry: QueryRegistry, query_type: Type[Q]) -> None:
        handler = query_registry.get_query_destination(query_type)
        self.__server.register(query_type, handler)

    def start(self) -> None:
        self.__server.run()

    def stop(self) -> None:
        pass



from typing import Type

from bus_station.passengers.reception.passenger_receiver import PassengerReceiver
from bus_station.passengers.serialization.passenger_deserializer import PassengerDeserializer
from bus_station.query_terminal.query import Query
from bus_station.query_terminal.query_handler import QueryHandler
from bus_station.query_terminal.serialization.query_response_serializer import QueryResponseSerializer
from bus_station.shared_terminal.rpyc_server import RPyCServer


class RPyCQueryServer(RPyCServer[Query, QueryHandler]):
    def __init__(
        self,
        host: str,
        port: int,
        query_deserializer: PassengerDeserializer,
        query_receiver: PassengerReceiver[Query, QueryHandler],
        query_response_serializer: QueryResponseSerializer,
    ):
        super().__init__(host, port, query_deserializer, query_receiver)
        self.__query_response_serializer = query_response_serializer

    def _passenger_handler(
        self, bus_stop: QueryHandler, passenger_class: Type[Query], serialized_passenger: str
    ) -> str:
        query = self._passenger_deserializer.deserialize(serialized_passenger, passenger_cls=passenger_class)
        if not isinstance(query, Query):
            raise TypeError("Input serialized query is not a Query")
        query_response = self._passenger_receiver.receive(query, bus_stop)
        return self.__query_response_serializer.serialize(query_response)



from datetime import datetime
from typing import Optional, Type

from bus_station.tracking_terminal.models.passenger_tracking import PassengerTracking


class MongoPassengerTrackingDeserializer:
    def deserialize(
        self, serialized_tracking_data: dict, target_tracking_class: Type[PassengerTracking]
    ) -> PassengerTracking:
        id_value = serialized_tracking_data.pop("_id")
        serialized_tracking_data["id"] = id_value
        serialized_tracking_data["execution_start"] = self.__deserialize_timestamp(
            serialized_tracking_data["execution_start"]
        )
        serialized_tracking_data["execution_end"] = self.__deserialize_timestamp(
            serialized_tracking_data["execution_end"]
        )
        return target_tracking_class(**serialized_tracking_data)

    def __deserialize_timestamp(self, timestamp: Optional[float]) -> Optional[datetime]:
        if timestamp is None:
            return None
        return datetime.fromtimestamp(timestamp)



from abc import ABC, abstractmethod
from typing import Optional, Type

from bus_station.query_terminal.query import Query
from bus_station.query_terminal.query_handler import QueryHandler
from bus_station.query_terminal.registry.query_registry import QueryRegistry


class RemoteQueryRegistry(QueryRegistry, ABC):
    @abstractmethod
    def _register(self, query: Type[Query], handler: QueryHandler, handler_contact: str) -> None:
        pass

    @abstractmethod
    def get_query_destination_contact(self, query: Type[Query]) -> Optional[str]:
        pass



from abc import abstractmethod
from typing import Optional, Protocol, Type

from bus_station.passengers.passenger import Passenger


class PassengerDeserializer(Protocol):
    @abstractmethod
    def deserialize(self, passenger_serialized: str, passenger_cls: Optional[Type[Passenger]] = None) -> Passenger:
        pass



from dataclasses import asdict
from datetime import datetime
from typing import Optional

from bus_station.tracking_terminal.models.passenger_tracking import PassengerTracking


class MongoPassengerTrackingSerializer:
    def serialize(self, tracking_instance: PassengerTracking) -> dict:
        instance_data = asdict(tracking_instance)
        id_value = instance_data.pop("id")
        instance_data["_id"] = id_value
        instance_data["execution_start"] = self.__serialize_date_time(tracking_instance.execution_start)
        instance_data["execution_end"] = self.__serialize_date_time(tracking_instance.execution_end)
        return instance_data

    def __serialize_date_time(self, date_time: Optional[datetime]) -> Optional[float]:
        if date_time is None:
            return None
        return date_time.timestamp()



import time
from logging import Logger
from typing import Optional

from bus_station.query_terminal.middleware.query_middleware import QueryMiddleware
from bus_station.query_terminal.query import Query
from bus_station.query_terminal.query_handler import QueryHandler
from bus_station.query_terminal.query_response import QueryResponse


class TimingQueryMiddleware(QueryMiddleware):
    def __init__(self, logger: Logger):
        self.__logger = logger
        self.__start_time = None

    def before_handle(self, passenger: Query, bus_stop: QueryHandler) -> None:
        self.__start_time = time.time()

    def after_handle(
        self,
        passenger: Query,
        bus_stop: QueryHandler,
        query_response: QueryResponse,
        handling_exception: Optional[Exception] = None,
    ) -> QueryResponse:
        execution_time = time.time() - self.__start_time
        execution_result_hint = "successfully" if handling_exception is None else "wrongly"
        self.__logger.info(
            f"Query {passenger} handled {execution_result_hint} "
            f"by {bus_stop.bus_stop_name()} in {execution_time} seconds"
        )
        return query_response



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['bus_station',
 'bus_station.command_terminal',
 'bus_station.command_terminal.bus',
 'bus_station.command_terminal.bus.asynchronous',
 'bus_station.command_terminal.bus.asynchronous.distributed',
 'bus_station.command_terminal.bus.synchronous',
 'bus_station.command_terminal.bus.synchronous.distributed',
 'bus_station.command_terminal.bus_engine',
 'bus_station.command_terminal.middleware',
 'bus_station.command_terminal.middleware.implementations',
 'bus_station.command_terminal.registry',
 'bus_station.event_terminal',
 'bus_station.event_terminal.bus',
 'bus_station.event_terminal.bus.asynchronous',
 'bus_station.event_terminal.bus.asynchronous.distributed',
 'bus_station.event_terminal.bus.synchronous',
 'bus_station.event_terminal.bus_engine',
 'bus_station.event_terminal.middleware',
 'bus_station.event_terminal.middleware.implementations',
 'bus_station.event_terminal.registry',
 'bus_station.passengers',
 'bus_station.passengers.passenger_record',
 'bus_station.passengers.reception',
 'bus_station.passengers.serialization',
 'bus_station.query_terminal',
 'bus_station.query_terminal.bus',
 'bus_station.query_terminal.bus.synchronous',
 'bus_station.query_terminal.bus.synchronous.distributed',
 'bus_station.query_terminal.bus_engine',
 'bus_station.query_terminal.middleware',
 'bus_station.query_terminal.middleware.implementations',
 'bus_station.query_terminal.registry',
 'bus_station.query_terminal.serialization',
 'bus_station.shared_terminal',
 'bus_station.shared_terminal.broker_connection',
 'bus_station.shared_terminal.broker_connection.connection_parameters',
 'bus_station.shared_terminal.bus_stop_resolver',
 'bus_station.shared_terminal.engine',
 'bus_station.shared_terminal.engine.runner',
 'bus_station.shared_terminal.factories',
 'bus_station.tracking_terminal',
 'bus_station.tracking_terminal.mappers',
 'bus_station.tracking_terminal.mappers.sqlalchemy',
 'bus_station.tracking_terminal.models',
 'bus_station.tracking_terminal.repositories',
 'bus_station.tracking_terminal.repositories.implementations',
 'bus_station.tracking_terminal.repositories.implementations.pymongo',
 'bus_station.tracking_terminal.repositories.implementations.sqlalchemy',
 'bus_station.tracking_terminal.trackers']

package_data = \
{'': ['*']}

install_requires = \
['SQLAlchemy==1.4.29',
 'jsonrpcclient>=4.0.2,<5.0.0',
 'jsonrpcserver>=5.0.6,<6.0.0',
 'kombu==5.2.2',
 'pymongo==4.0.1',
 'pypendency>=0.1.0,<0.2.0',
 'redis==4.1.0',
 'requests>=2.27.1,<3.0.0',
 'rpyc==5.0.1']

setup_kwargs = {
    'name': 'bus-station',
    'version': '3.0.2',
    'description': 'A python bus station',
    'long_description': '# Bus station\nPython bus pattern implementation\n',
    'author': 'DeejayRevok',
    'author_email': 'seryi_one@hotmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.10.0,<3.11.0',
}


setup(**setup_kwargs)



from ctypes import c_bool
from multiprocessing import Value
from typing import Callable, List, Type

from kombu import Connection, Message
from kombu.messaging import Consumer as KombuConsumer
from kombu.messaging import Queue
from kombu.mixins import ConsumerMixin
from kombu.transport.virtual import Channel

from bus_station.passengers.passenger import Passenger
from bus_station.passengers.reception.passenger_receiver import PassengerReceiver
from bus_station.passengers.serialization.passenger_deserializer import PassengerDeserializer
from bus_station.shared_terminal.bus_stop import BusStop


class PassengerKombuConsumer(ConsumerMixin):
    def __init__(
        self,
        connection: Connection,
        passenger_queue: Queue,
        passenger_bus_stop: BusStop,
        passenger_class: Type[Passenger],
        passenger_receiver: PassengerReceiver,
        passenger_deserializer: PassengerDeserializer,
    ):
        self.connection = connection
        self.__should_stop = Value(c_bool, False)
        self.__queue = passenger_queue
        self.__passenger_bus_stop = passenger_bus_stop
        self.__passenger_class = passenger_class
        self.__passenger_receiver = passenger_receiver
        self.__passenger_deserializer = passenger_deserializer

    @property
    def should_stop(self):
        return self.__should_stop.value

    @should_stop.setter
    def should_stop(self, value):
        self.__should_stop.value = value

    def get_consumers(self, Consumer: Callable, channel: Channel) -> List[KombuConsumer]:
        return [
            Consumer(queues=[self.__queue], callbacks=[self.on_message]),
        ]

    def on_message(self, body: str, message: Message) -> None:
        passenger = self.__passenger_deserializer.deserialize(body, passenger_cls=self.__passenger_class)
        try:
            self.__passenger_receiver.receive(passenger, self.__passenger_bus_stop)
            message.ack()
        except Exception:
            message.reject(requeue=False)

    def stop(self):
        self.should_stop = True



from abc import abstractmethod

from bus_station.event_terminal.event import Event
from bus_station.shared_terminal.bus import Bus


class EventBus(Bus[Event]):
    @abstractmethod
    def transport(self, passenger: Event) -> None:
        pass



from abc import abstractmethod
from typing import Iterable, List, Optional, Protocol

from bus_station.passengers.passenger_record.passenger_record import PassengerRecord


class PassengerRecordRepository(Protocol):
    @abstractmethod
    def save(self, passenger_record: PassengerRecord) -> None:
        pass

    @abstractmethod
    def find_by_passenger_name(self, passenger_name: str) -> Optional[List[PassengerRecord]]:
        pass

    @abstractmethod
    def find_by_passenger_name_and_destination(
        self, passenger_name: str, passenger_destination_fqn: str
    ) -> Optional[PassengerRecord]:
        pass

    @abstractmethod
    def all(self) -> Iterable[PassengerRecord]:
        pass

    @abstractmethod
    def delete_by_passenger_name(self, passenger_name: str) -> None:
        pass



from typing import Optional, Type, TypeVar

from bus_station.command_terminal.command import Command
from bus_station.command_terminal.registry.command_registry import CommandRegistry
from bus_station.shared_terminal.engine.engine import Engine
from bus_station.shared_terminal.rpyc_server import RPyCServer

C = TypeVar("C", bound=Command)


class RPyCCommandBusEngine(Engine):
    def __init__(
        self, rpyc_server: RPyCServer, command_registry: CommandRegistry, command_type: Optional[Type[C]] = None
    ):
        super().__init__()
        self.__server = rpyc_server

        if command_type is not None:
            self.__register_command_in_server(command_registry, command_type)
            return

        for command_type in command_registry.get_commands_registered():
            self.__register_command_in_server(command_registry, command_type)

    def __register_command_in_server(self, command_registry: CommandRegistry, command_type: Type[C]) -> None:
        handler = command_registry.get_command_destination(command_type)
        self.__server.register(command_type, handler)

    def start(self) -> None:
        self.__server.run()

    def stop(self) -> None:
        pass



from multiprocessing import Queue

from bus_station.command_terminal.bus.command_bus import CommandBus
from bus_station.command_terminal.command import Command
from bus_station.command_terminal.handler_not_found_for_command import HandlerNotFoundForCommand
from bus_station.command_terminal.registry.in_memory_command_registry import InMemoryCommandRegistry
from bus_station.passengers.serialization.passenger_serializer import PassengerSerializer


class MemoryQueueCommandBus(CommandBus):
    def __init__(
        self,
        command_serializer: PassengerSerializer,
        command_registry: InMemoryCommandRegistry,
    ):
        self.__command_serializer = command_serializer
        self.__command_registry = command_registry

    def transport(self, passenger: Command) -> None:
        handler_queue = self.__command_registry.get_command_destination_contact(passenger.__class__)
        if handler_queue is None:
            raise HandlerNotFoundForCommand(passenger.passenger_name())

        self.__put_command(handler_queue, passenger)

    def __put_command(self, queue: Queue, command: Command) -> None:
        serialized_command = self.__command_serializer.serialize(command)
        queue.put(serialized_command)



from dataclasses import MISSING, Field, fields
from importlib import import_module
from json import loads
from typing import Generic, Optional, Type, TypeVar

from bus_station.passengers.passenger import Passenger
from bus_station.passengers.serialization.passenger_deserialization_error import PassengerDeserializationError
from bus_station.passengers.serialization.passenger_deserializer import PassengerDeserializer

P = TypeVar("P", bound=Passenger)


class PassengerJSONDeserializer(PassengerDeserializer, Generic[P]):
    def deserialize(self, passenger_serialized: str, passenger_cls: Optional[Type[P]] = None) -> P:
        deserialized_data = loads(passenger_serialized)
        if passenger_cls is None:
            passenger_cls = self.__get_passenger_dataclass(deserialized_data)
        return self.__from_passenger_cls(passenger_cls, deserialized_data["passenger_data"])

    def __get_passenger_dataclass(self, passenger_dict: dict) -> Type[P]:
        passenger_type = passenger_dict["passenger_type"]
        passenger_module_name, passenger_class_name = passenger_type.rsplit(".", 1)
        return getattr(import_module(passenger_module_name), passenger_class_name)

    def __from_passenger_cls(self, passenger_cls: Type[P], passenger_data: dict) -> P:
        passenger_cls_field_values = dict()
        for passenger_field in fields(passenger_cls):
            passenger_field: Field = passenger_field
            if passenger_field.default is MISSING and passenger_field.name not in passenger_data:
                raise PassengerDeserializationError(passenger_cls, f"Missing value for field {passenger_field.name}")
            if passenger_field.name not in passenger_data:
                continue
            passenger_cls_field_values[passenger_field.name] = passenger_data[passenger_field.name]
        return passenger_cls(**passenger_cls_field_values)



import requests
from jsonrpcclient import request
from jsonrpcclient.responses import Error, to_result

from bus_station.command_terminal.bus.command_bus import CommandBus
from bus_station.command_terminal.command import Command
from bus_station.command_terminal.command_execution_failed import CommandExecutionFailed
from bus_station.command_terminal.handler_not_found_for_command import HandlerNotFoundForCommand
from bus_station.command_terminal.registry.remote_command_registry import RemoteCommandRegistry
from bus_station.passengers.serialization.passenger_serializer import PassengerSerializer


class JsonRPCCommandBus(CommandBus):
    def __init__(
        self,
        command_serializer: PassengerSerializer,
        command_registry: RemoteCommandRegistry,
    ):
        self.__command_serializer = command_serializer
        self.__command_registry = command_registry

    def transport(self, passenger: Command) -> None:
        handler_address = self.__command_registry.get_command_destination_contact(passenger.__class__)
        if handler_address is None:
            raise HandlerNotFoundForCommand(passenger.passenger_name())

        self.__execute_command(passenger, handler_address)

    def __execute_command(self, command: Command, command_handler_addr: str) -> None:
        serialized_command = self.__command_serializer.serialize(command)

        request_response = requests.post(
            command_handler_addr, json=request(command.__class__.passenger_name(), params=(serialized_command,))
        )
        json_rpc_response = to_result(request_response.json())

        if isinstance(json_rpc_response, Error):
            raise CommandExecutionFailed(command, json_rpc_response.message)



from abc import ABC, abstractmethod

from bus_station.shared_terminal.engine.engine import Engine


class EngineRunner(ABC):
    def __init__(self, engine: Engine):
        self._engine = engine

    def __enter__(self) -> None:
        self.run()

    def __exit__(self, _, __, ___) -> None:
        self.stop()

    @abstractmethod
    def run(self) -> None:
        pass

    @abstractmethod
    def stop(self) -> None:
        pass



from kombu import Connection
from kombu.messaging import Producer

from bus_station.command_terminal.bus.command_bus import CommandBus
from bus_station.command_terminal.command import Command
from bus_station.command_terminal.handler_not_found_for_command import HandlerNotFoundForCommand
from bus_station.command_terminal.registry.remote_command_registry import RemoteCommandRegistry
from bus_station.passengers.serialization.passenger_serializer import PassengerSerializer


class KombuCommandBus(CommandBus):
    def __init__(
        self,
        broker_connection: Connection,
        command_serializer: PassengerSerializer,
        command_registry: RemoteCommandRegistry,
    ):
        self.__broker_connection = broker_connection
        self.__command_serializer = command_serializer
        self.__command_registry = command_registry
        self.__producer = Producer(broker_connection.channel())

    def transport(self, passenger: Command) -> None:
        handler_queue_name = self.__command_registry.get_command_destination_contact(passenger.__class__)
        if handler_queue_name is None:
            raise HandlerNotFoundForCommand(passenger.passenger_name())

        self.__publish_command(passenger, handler_queue_name)

    def __publish_command(self, command: Command, routing_key: str) -> None:
        serialized_command = self.__command_serializer.serialize(command)
        if self.__producer is not None:
            self.__producer.publish(
                serialized_command,
                exchange="",
                routing_key=routing_key,
                retry=True,
                retry_policy={
                    "interval_start": 0,
                    "interval_step": 2,
                    "interval_max": 10,
                    "max_retries": 10,
                },
            )

    def shutdown(self) -> None:
        self.__producer.release()



from typing import Dict, Optional, TypeVar

from bus_station.shared_terminal.bus_stop import BusStop
from bus_station.shared_terminal.bus_stop_resolver.bus_stop_resolver import BusStopResolver
from bus_station.shared_terminal.fqn_getter import FQNGetter

S = TypeVar("S", bound=BusStop)


class InMemoryBusStopResolver(BusStopResolver[S]):
    def __init__(self, fqn_getter: FQNGetter):
        self.__fqn_getter = fqn_getter
        self.__bus_stops: Dict[str, S] = {}

    def add_bus_stop(self, bus_stop: S) -> None:
        fqn = self.__fqn_getter.get(bus_stop)
        self.__bus_stops[fqn] = bus_stop

    def resolve_from_fqn(self, bus_stop_fqn: str) -> Optional[S]:
        return self.__bus_stops.get(bus_stop_fqn)



from kombu import Connection
from kombu.messaging import Producer

from bus_station.event_terminal.bus.event_bus import EventBus
from bus_station.event_terminal.event import Event
from bus_station.event_terminal.registry.remote_event_registry import RemoteEventRegistry
from bus_station.passengers.serialization.passenger_serializer import PassengerSerializer


class KombuEventBus(EventBus):
    def __init__(
        self,
        broker_connection: Connection,
        event_serializer: PassengerSerializer,
        event_registry: RemoteEventRegistry,
    ):
        self.__broker_connection = broker_connection
        self.__event_serializer = event_serializer
        self.__event_registry = event_registry
        self.__producer = Producer(broker_connection.channel())

    def transport(self, passenger: Event) -> None:
        event_exchange_names = self.__event_registry.get_event_destination_contacts(passenger.__class__)
        if event_exchange_names is None:
            return

        published_exchanges = []
        for event_exchange_name in event_exchange_names:
            if event_exchange_name in published_exchanges:
                continue
            self.__publish_event(passenger, event_exchange_name)
            published_exchanges.append(event_exchange_name)

    def __publish_event(self, event: Event, event_exchange_name: str) -> None:
        serialized_event = self.__event_serializer.serialize(event)
        if self.__producer is not None:
            self.__producer.publish(
                serialized_event,
                exchange=event_exchange_name,
                retry=True,
                retry_policy={
                    "interval_start": 0,
                    "interval_step": 2,
                    "interval_max": 10,
                    "max_retries": 10,
                },
            )

    def shutdown(self) -> None:
        self.__producer.release()



from typing import ClassVar, Union, final
from urllib.parse import urlencode

from bus_station.shared_terminal.broker_connection.connection_parameters.connection_parameters import (
    ConnectionParameters,
)


@final
class RabbitMQConnectionParameters(ConnectionParameters):

    __CONNECTION_STRING_PATTERN: ClassVar[str] = "amqp://{username}:{password}@{host}:{port}/{vhost}"

    def __init__(self, host: str, port: Union[int, str], username: str, password: str, vhost: str, **additional_args):
        self.__host = host
        self.__port = port
        self.__username = username
        self.__password = password
        self.__vhost = vhost
        self.__additional_args = additional_args

    def get_connection_string(self) -> str:
        base_connection_str = self.__CONNECTION_STRING_PATTERN.format(
            host=self.__host,
            port=self.__port,
            username=self.__username,
            password=self.__password,
            vhost=self.__vhost,
        )
        if self.__additional_args is None or len(self.__additional_args) == 0:
            return base_connection_str

        base_connection_str += "?"
        base_connection_str += urlencode(self.__additional_args)
        return base_connection_str



import time
from logging import Logger
from typing import Optional

from bus_station.command_terminal.command import Command
from bus_station.command_terminal.command_handler import CommandHandler
from bus_station.command_terminal.middleware.command_middleware import CommandMiddleware


class TimingCommandMiddleware(CommandMiddleware):
    def __init__(self, logger: Logger):
        self.__logger = logger
        self.__start_time = None

    def before_handle(self, passenger: Command, bus_stop: CommandHandler) -> None:
        self.__start_time = time.time()

    def after_handle(
        self, passenger: Command, bus_stop: CommandHandler, handling_exception: Optional[Exception] = None
    ) -> None:
        execution_time = time.time() - self.__start_time
        execution_result_hint = "successfully" if handling_exception is None else "wrongly"
        self.__logger.info(
            f"Command {passenger} handled {execution_result_hint} "
            f"by {bus_stop.bus_stop_name()} in {execution_time} seconds"
        )



from rpyc import Connection, connect

from bus_station.command_terminal.bus.command_bus import CommandBus
from bus_station.command_terminal.command import Command
from bus_station.command_terminal.handler_not_found_for_command import HandlerNotFoundForCommand
from bus_station.command_terminal.registry.remote_command_registry import RemoteCommandRegistry
from bus_station.passengers.serialization.passenger_serializer import PassengerSerializer


class RPyCCommandBus(CommandBus):
    def __init__(
        self,
        command_serializer: PassengerSerializer,
        command_registry: RemoteCommandRegistry,
    ):
        self.__command_serializer = command_serializer
        self.__command_registry = command_registry

    def transport(self, passenger: Command) -> None:
        handler_address = self.__command_registry.get_command_destination_contact(passenger.__class__)
        if handler_address is None:
            raise HandlerNotFoundForCommand(passenger.passenger_name())

        rpyc_client = self.__get_rpyc_client(handler_address)
        self.__execute_command(rpyc_client, passenger)
        rpyc_client.close()

    def __get_rpyc_client(self, handler_addr: str) -> Connection:
        host, port = handler_addr.split(":")
        return connect(host, port=port, config={"allow_public_attrs": True})

    def __execute_command(self, client: Connection, command: Command) -> None:
        serialized_command = self.__command_serializer.serialize(command)
        getattr(client.root, command.passenger_name())(serialized_command)



from typing import Type

from bus_station.passengers.passenger import Passenger
from bus_station.tracking_terminal.mappers.sqlalchemy.sqlalchemy_mapper import SQLAlchemyMapper
from bus_station.tracking_terminal.models.passenger_tracking import PassengerTracking
from bus_station.tracking_terminal.repositories.passenger_tracking_repository import PassengerTrackingRepository
from bus_station.tracking_terminal.trackers.passenger_tracker import PassengerTracker


class SQLAlchemyPassengerTracker(PassengerTracker):
    def __init__(self, passenger_tracking_repository: PassengerTrackingRepository, sqlalchemy_mapper: SQLAlchemyMapper):
        super().__init__(passenger_tracking_repository)
        self.__sqlalchemy_mapper = sqlalchemy_mapper

    def _get_tracking_model(self, passenger: Passenger) -> Type:
        tracking_model = self.__sqlalchemy_mapper.model
        if tracking_model != PassengerTracking and not issubclass(tracking_model, PassengerTracking):
            raise TypeError("Mapped tracking model is not an instance of PassengerTracking")
        return tracking_model
